extern long int precise_time (void);
extern void info_arc (LLINT * xc, LLINT * yc, int *theta1, int *theta2,
		      int *dir,
		      double x1, double y1, double x2, double y2,
		      double x3, double y3);
extern void info_arc_export (double *xc, double *yc, double *theta1,
                             double *theta2, int *dir, double x1, double y1,
			     double x2, double y2, double x3, double y3);
extern double dist2_to_arc (struct Bond *ref, const LLINT x,
                	    const LLINT y);
extern int read_config_file (void);
extern int write_config_file (void);
