#define U2S(x) ((int) ((((double) x) + UNITS / 2) / UNITS * zoom))
#define S2U(x) ((int) ((x) * UNITS / zoom))

void adjust_atom (struct Bond *atom, GtkWidget * area, const int mode,
                  const double zoom);
void adjust_all_atom_in_list (struct Bond *list, GtkWidget * area,
                              const int mode, const double zoom);

void draw_ornaments (struct Ornament *orn, const LLINT x, const LLINT y,
                     const double zoom, GdkGC *gc, GdkPixmap *pix);
void draw_bond_list (struct Bond *list, const int recursive,
                     const int selected, GtkWidget * area, GdkPixmap *pix,
		     const int mode, const double zoom);
