enum { PROGRAM_UNKNOWN, PROGRAM_NOT_FOUND, PROGRAM_IN_PATH,
       PROGRAM_IN_USR_BIN, PROGRAM_IN_USR_LOCAL_BIN,
       PROGRAM_IN_PROVIDED_PATH, PROGRAM_IN_PREF_PATH, PROGRAM_IN_SW_BIN };

extern int detect_gs (void);
extern int detect_pstoedit (void);
