extern gchar *edit_text (const gchar * text, unsigned int *type,
                         double *size, GdkColor *color, GtkWindow *parent);
extern void edit_bond (struct Bond *bond, GtkWindow * parent);
extern void edit_ornaments (struct Bond *list, struct Bond *bond,
	                    const unsigned int end, GtkWindow * parent);
extern void prop_dialog (void);

#ifdef MACOSX
extern void macosx_noxrunning (void);
extern void macosx_initializecarbon (void);
#endif
