extern void eps_export (struct Bond *list, char * filename,
			GtkWindow * parent, int try_gs);
extern void fig_export (struct Bond *list, char * filename,
	                GtkWindow * parent);
extern void fig_pstoedit_export (struct Bond *list, char * filename,
	                GtkWindow * parent);
extern void pdf_export (struct Bond *list, char * filename,
	                GtkWindow * parent);
extern void eps_size_export (struct Bond *list, char * filename,
	                     GtkWindow * parent);
